# VerChor

This document is intended to the users of VerChor software.
It provides instructions about how to install and run VerChor in your own machine. 

## Installation

### Java RE

This is needed to transform BPMN files into the CIF format.
It is not required if you plan to use the CIF format directly.

VerChor has been tested with Java RE 1.7.

To install this, you can go to web site: http://www.java.com/fr/download/index.jsp.

### Eclipse and BPMN modeling plugin

1. Download Eclipse from its official web site: http://www.eclipse.org/downloads/.

      Eclipse Kepler 4.3 is recommended, Eclipse Indigo 3.7 and Eclipse Juno 4.2 have also been tested.

2. Install the BPMN2 Modeler plugin into your Eclipse

   a. Start Eclipse, then select `Help > Install New Software`

   b. Click Add, in the top-right corner

   c. In the Add Repository dialog that appears, enter "BPMN2-Modeler" for the name and the URL that you have chosen from the seb site (e.g., http://download.eclipse.org/bpmn2-modeler/updates/kepler/1.0.1/)

   d. Choose the item appearing in the name box and then click next

   e. Click next and then choose to accept the terms of license before clicking finish

   f. If during the installation, a "security warning" dialog pops up, click OK

   g. When the installation completes, restart Eclipse.

It is possible that some versions cannot work in your machine, in which case you can try other versions.

### Python and PyXB

The software is mainly written in Python, so it is necessary to install Python.
Python 2.7.* version have been tested but Python3.* is not compatible.

1. Download python from the web site: https://www.python.org/download/releases/2.7/.

2. Download PyXB library for XML in python: https://sourceforge.net/projects/pyxb/files/pyxb/

      The 1.1.* release series should work (tested with version 1.1.1). Newer versions are know to be incompatible with VerChor.

      Once you have the library archive, open it and run:
      ```sh
      python setup.py install
      ```
### CADP

VerChor reuses CADP toolbox to automate some key analysis tasks using model and equivalence checking.
Thus, to run VerChor, you need to install CADP beforehand.

To do this, please go to CADP registration web site: http://cadp.inria.fr/registration/.

There, you have to fill in the registration form and then submit it.
You will receive an email very soon to provide you precise instructions about how to get and install the latest version of CADP in your machine, including the password required during the installation.

## Setup of the framework

### Copying the framework files

If you are reading this, you already have opened the zip archive.
You then have to copy these in your Eclipse workspace folder:

- subfolder `src/py`
- subfolder `src/java`
- subfolder `src/xml`

You also have to update the environment variables to your own ones in file `src/py/mainEclipse.py`

You can also copy the provided examples there if you want:

- subfolder `ex`

### Generating the bindings for the CIF format (optional)

After installing PyXB, the local bindings can be generated by the following command:

```sh
pyxbgen -u choreoFI.xsd -m choreo_xml_model
```

This will create the `choreo_xml_model.py` file from the XSD schema definition.
This file is used by the other scripts to read generated CIF files.
Initially, the `choreo_xml_model.py` file is already provided in the downloaded package. 

### Setting up Eclipse external tools

You have to set up five Eclipse external tools that correspond to the VerChor scripts in order to work on choreographies directly from your Eclipse IDE. This includes checking properties (synchronizability, realizability and repairability) and automatically repairing a BPMN choreography if realizability is not satisfied.

To install a new external tool, select `Run > External Tools > External tools configurations` and then fill in the configuration dialog.

In the following, an example of external tool configuration for each external tool is given with the hypothesis that your workspace folder is WORKSPACE, that the java command is JAVA (e.g., usr/bin/java) and that the python command is PYTHON (e.g., /usr/bin/python).

1. BPMN to CIF transformation

      - Name: Generate Intermediate Format
      - Location: JAVA (e.g., /usr/bin/java)
      - Working Directory: WORKSPACE
      - Arguments:

            ```
            -cp "WORKSPACE/src/java/fmt.jar:WORKSPACE/eclipse/plugins/*" transformations.bpmn2cif.Bpmn2Cif
            ${selected_resource_loc}
            ```

2. Generation of the LTSs used to check properties

      - Name: Generate LTS
      - Location: PYTHON (e.g., /usr/bin/python)
      - Working Directory: WORKSPACE
      - Arguments:

            ```
            WORKSPACE/src/py/mainEclipse.py
            ${selected_resource_loc}
            LTSgeneration
            ```

3. Verification of synchronizability and realizability

      - Name: Check Synchronizability and Realizability
      - Location: PYTHON (e.g., /usr/bin/python)
      - Working Directory: WORKSPACE
      - Arguments:

            ```
            WORKSPACE/src/py/mainEclipse.py
            ${selected_resource_loc}
            SynchronizabilityRealizability
            ```

4. Generation of a controller to fix the choreography if it is not realizable

      - Name: Check Repairability and Generate Controller
      - Location: PYTHON (e.g., /usr/bin/python)
      - Working Directory: WORKSPACE
      - Arguments:

            ```
            WORKSPACE/src/py/mainEclipse.py
            ${selected_resource_loc}
            Controlgeneration
            ```

5. Clean up of generated files

      - Name: Clean Up
      - Location: PYTHON (e.g., /usr/bin/python)
      - Working Directory: WORKSPACE
      - Arguments:

            ```
            WORKSPACE/src/py/clean.py
            ${selected_resource_loc}
            ```

## Usage

The following steps demonstrate how to use VerChor on an example.
The order of the steps should be strictly respected.

1. Create a new general project in Eclipse.

2. From this project, import a BPMN choreography in the example folder, (e.g. `NonSynchro.bpmn`).

      You can also create your own BPMN choreography using the BPMN modeling plugin.

3. Choose the selected BPMN file and then select `Run > External Tools > Generate Intermediate Format`.

      The CIF file for the choreography is generated.

4. Choose the same BPMN file and then select `Run > External Tools > Generate LTS`.

      Now the corresponding LTSs used for property checking are generated.
      Some new `.bcg` files will appear in your workspace folder.
      You can see a generated LTS using the ```bcg_edit filename.bcg``` shell command.

5. Choose the same BPMN file and then select `Run > External Tools > Check Synchronizability and Realizability`.

      This checks the synchronizability and realizability of the choreography.
      You will see the result in the Eclipse console.
      If the choreography is not realizable, you can continue to step 6.

6. Choose the same BPMN file and then select `Run > External Tools > Check Repairability and Generate Controller`.

      This is used to fix a non realizable choreography, if it is repairable, in such a way that it becomes synchronizable and realizable.
      More precisely, if the choreography is not repairable, the program will output a `non repairable` message and then stop. Otherwise, the program will continue to repair the choreography through one or more steps until the  choreography is synchronizable and realizable.
      The choreography shown at the last step is a synchronizable and realizable one.

      If you want to fix the choreography on your own, you can modify the choreography directly in the modeler and then begin from step 3 until step 5 to see whether the new version is synchronizable and realizable.

7. Choose the same BPMN file and then select `Run > External Tools > Clean Up`.

      Files generated in the previous steps are removed.
